﻿using FT.Architecture.Exceptions;
using FT.Architecture.NHibernate.Factories;
using FT.Architecture.NHibernate.SessionManagement;
using FT.Architecture.UnitOfWork;
using NHibernate;

namespace FT.Architecture.NHibernate.Context
{
    public class NHibernateUnitOfWork : IUnitOfWork
    {
        private readonly ISessionProviderService _sessionProviderService;
        private ISessionFactory _factory;
        private static NHibernateSessionManager _sessionManager;

        public NHibernateUnitOfWork(ISessionProviderService sessionProviderService, INHibernateConfigurator nHibernateConfigurator)
        {
            _sessionProviderService = sessionProviderService;
            if (_sessionManager == null)
                _sessionManager = NHibernateSessionManager.GetInstance(nHibernateConfigurator);
        }

        public void Initialise()
        {
            if (_sessionProviderService == null)
            {
                throw new UnitOfWorkException("The session provider service is not initialised", this);
            }

            _sessionProviderService.SetSession(null);
            _factory = _sessionManager.GetSessionFactoryFor("sf");
        }

        public IUnitOfWorkInstance Start()
        {
            if (_factory == null)
            {
                Initialise();
            }

            if (_factory == null)
            {
                throw new UnitOfWorkException("The NHibernate session factory is not initialised", this);
            }

            ISession session = _factory.OpenSession();
            
            if (session == null)
            {
                throw new UnitOfWorkException("The NHibernate session is not initialised", this);
            }

            session.BeginTransaction();

            if (_sessionProviderService == null)
            {
                throw new UnitOfWorkException("The session provider service is not initialised", this);
            }

            _sessionProviderService.SetSession(session);

            return new UnitOfWorkInstance(this);
        }

        public void Commit()
        {
            if (_sessionProviderService == null)
            {
                throw new UnitOfWorkException("The session provider service is not initialised", this);
            }

            var session = _sessionProviderService.GetSession();

            if (session != null)
            {
                ITransaction transaction = session.Transaction;
                if (transaction.IsActive)
                {
                    transaction.Commit();
                }
                session.Dispose();
                _sessionProviderService.SetSession(null);
            }
        }

        public void Rollback()
        {
            if (_sessionProviderService == null)
            {
                throw new UnitOfWorkException("The session provider service is not initialised", this);
            }

            var session = _sessionProviderService.GetSession();

            if (session != null)
            {
                ITransaction transaction = session.Transaction;
                if (transaction.IsActive)
                {
                    transaction.Rollback();
                }
                session.Dispose();
                _sessionProviderService.SetSession(null);
            }
        }

        public void Flush()
        {
            if (_sessionProviderService == null)
            {
                throw new UnitOfWorkException("The session provider service is not initialised", this);
            }

            var session = _sessionProviderService.GetSession();

            if (session != null)
            {
                session.Flush();
            }
        }

        public bool IsStarted()
        {
            if (!_sessionProviderService.HasSession())
                return false;

            var session = _sessionProviderService.GetSession();
            return session.IsOpen;
        }
    }
}
