#region Copyright
//       Copyright (c) 2009 - Antoine Jaussoin - http://www.jaussoin.com/blog/
// 
//        This program is free software: you can redistribute it and/or modify
//        it under the terms of the GNU General Public License as published by
//        the Free Software Foundation, either version 3 of the License, or
//        (at your option) any later version.
// 
//        This program is distributed in the hope that it will be useful,
//        but WITHOUT ANY WARRANTY; without even the implied warranty of
//        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//        GNU General Public License for more details.
// 
//        You should have received a copy of the GNU General Public License
//        along with this program.  If not, see <http://www.gnu.org/licenses/>.
#endregion
using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;

namespace FT.Architecture.Utils.Reflection
{
    /// <summary>
    /// Provide methods used in reflection
    /// </summary>
    public static class Reflection
    {
        /// <summary>
        /// Gets a member from an object
        /// The member name can be composed of other member name (such as "Statistics.Decile.Value")
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="object"></param>
        /// <param name="memberName"></param>
        /// <param name="ignoreComposedMemberNull">Don't throw an exception if an intermediate is <see langword="null"/>. For example, if looking for "Statistics.Decile.Value", Decile is <see langword="null"/>, then the method will return <see langword="null"/>.</param>
        /// <returns></returns>
        public static T GetProperty<T>(object @object, string memberName, bool ignoreComposedMemberNull)
        {
            if (@object == null)
                return default(T);

            string[] properties = memberName.Split('.');

            object current = @object;


            for (int i = 0; i < properties.Length; i++)
            {
                string prop = properties[i];
                object newCurrent = null;

                if (current == null)
                    if (ignoreComposedMemberNull)
                        return default(T);
                    else
                        throw new ReflectionPropertyParentIsNullException(prop, current);

                try
                {
                    PropertyInfo propertyInfo = current.GetType().GetProperty(prop);

                    if (propertyInfo == null)
                        throw new ReflectionPropertyDoesNotExistException(prop, current);

                    newCurrent = propertyInfo.GetValue(current, null);
                }
                catch (AmbiguousMatchException)
                {
                    foreach (PropertyInfo pi in current.GetType().GetProperties())
                    {
                        if (pi.Name == prop)
                        {
                            newCurrent = pi.GetValue(current, null);
                            break;
                        }
                    }
                }

                if (newCurrent == null && i < properties.Length - 1)
                    if (ignoreComposedMemberNull)
                        return default(T);
                    else
                        throw new ReflectionPropertyParentIsNullException(prop, current);

                current = newCurrent;
            }

            if (typeof (T) == typeof (string) && current != null)
                return (T) ((object) current.ToString());

            return (T) current;
        }

        /// <summary>
        /// Gets a member from an object
        /// The member name can be composed of other member name (such as "Statistics.Decile.Value")
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="object"></param>
        /// <param name="memberName"></param>
        /// <returns></returns>
        public static T GetProperty<T>(object @object, string memberName)
        {
            return GetProperty<T>(@object, memberName, false);
        }

        /// <summary>
        /// Gets a member from an object
        /// The member name can be composed of other member name (such as "Statistics.Decile.Value")
        /// </summary>
        /// <param name="object"></param>
        /// <param name="memberName"></param>
        /// <returns></returns>
        public static PropertyInfo GetProperty(object @object, string memberName)
        {
            string[] properties = memberName.Split('.');

            PropertyInfo currentPropertyInfo = null;
            object current = @object;
            for (int i = 0; i < properties.Length; i++)
            {
                string prop = properties[i];
                object currentValue = null;

                if (current == null)
                    throw new ReflectionPropertyParentIsNullException(prop, current);

                try
                {
                    currentPropertyInfo = current.GetType().GetProperty(prop);

                    if (currentPropertyInfo == null)
                        throw new ReflectionPropertyDoesNotExistException(prop, current);

                    currentValue = current.GetType().GetProperty(prop).GetValue(current, null);
                }
                catch (AmbiguousMatchException)
                {
                    foreach (PropertyInfo pi in current.GetType().GetProperties())
                    {
                        if (pi.Name == prop)
                        {
                            currentValue = pi.GetValue(current, null);
                            currentPropertyInfo = pi;
                            break;
                        }
                    }
                }

                if (currentValue == null && i < properties.Length - 1)
                    throw new ReflectionPropertyParentIsNullException(prop, current);

                current = currentValue;
            }

            return currentPropertyInfo;
        }

        /// <summary>
        /// Gets the value of a property of an object
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="object"></param>
        /// <param name="property"></param>
        /// <returns></returns>
        public static T GetProperty<T>(object @object, PropertyInfo property)
        {
            return (T) property.GetValue(@object, null);
        }

        /// <summary>
        /// Return all the properties of a given type of an object 
        /// </summary>
        /// <typeparam name="T">Type of the property</typeparam>
        /// <param name="object">Object to look on</param>
        /// <param name="recursive">Browse the object tree recursively to find all the occurrences of the object</param>
        /// <param name="alreadyBrowsed">Object already browsed (in order to avoid an infinite loop)</param>
        /// <returns>List of objects of type T</returns>
        private static List<T> GetProperties<T>(object @object, bool recursive, ICollection<object> alreadyBrowsed)
        {
            Type propertiesType = typeof (T);
            var properties = new List<T>();

            if (@object == null)
                return properties;

            if (!recursive)
            {
                foreach (PropertyInfo pi in @object.GetType().GetProperties())
                {
                    if (pi.PropertyType == propertiesType)
                        properties.Add((T) pi.GetValue(@object, null));
                }
            }
            else
            {
                throw new NotImplementedException("Recursivity Not implemented " + alreadyBrowsed.Count);
            }

            return properties;
        }

        /// <summary>
        /// Return all the properties of a given type of an object 
        /// </summary>
        /// <typeparam name="T">Type of the property</typeparam>
        /// <param name="object">Object to look on</param>
        /// <param name="recursive">Browse the object tree recursively to find all the occurrences of the object</param>
        /// <returns>List of objects of type T</returns>
        public static List<T> GetProperties<T>(object @object, bool recursive)
        {
            return GetProperties<T>(@object, recursive, new List<object>());
        }

        /// <summary>
        /// Return all the properties of a given type of an object 
        /// </summary>
        /// <typeparam name="T">Type of the property</typeparam>
        /// <param name="object">Object to look on</param>
        /// <returns>List of objects of type T</returns>
        public static List<T> GetProperties<T>(object @object)
        {
            return GetProperties<T>(@object, false);
        }

        /// <summary>
        /// Return all the fields of a given type of an object
        /// </summary>
        /// <typeparam name="T">Type of the field</typeparam>
        /// <param name="object">Object to look on</param>
        /// <returns>List of objects of type T</returns>
        public static List<T> GetFields<T>(object @object)
        {
            Type fieldsType = typeof (T);
            var fields = new List<T>();

            if (@object == null)
                return fields;

            foreach (FieldInfo pi in @object.GetType().GetFields())
            {
                if (pi.FieldType == fieldsType)
                    fields.Add((T) pi.GetValue(@object));
            }

            return fields;
        }

        /// <summary>
        /// Instantiate an object of the given type
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="typeToInstantiate"></param>
        /// <returns></returns>
        public static T Instantiate<T>(Type typeToInstantiate)
        {
            return (T) typeToInstantiate.Assembly.CreateInstance(typeToInstantiate.FullName);
        }

        /// <summary>
        /// Instantiate an object of the given type
        /// </summary>
        /// <param name="typeToInstantiate"></param>
        /// <returns></returns>
        public static object Instantiate(Type typeToInstantiate)
        {
            return Instantiate<object>(typeToInstantiate);
        }

        /// <summary>
        /// Get a <see cref="PropertyInfo"/> object from a lambda expression representing the member to return
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="expression"></param>
        /// <returns></returns>
        public static PropertyInfo GetProperty<T>(Expression<Func<T, object>> expression)
        {
            MemberExpression memberExpression = GetMemberExpression(expression);
            return (PropertyInfo)memberExpression.Member;
        }

        /// <summary>
        /// Get a string representation of the lambda expression.
        /// Example: a lambda of o => o.Item.SubItem will return a string of "Item.SubItem"
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="expression"></param>
        /// <returns></returns>
        public static string GetCombinedProperty<T>(this Expression<Func<T, object>> expression)
        {
            IEnumerable<PropertyInfo> properties = GetProperties(expression);
            properties = properties.Reverse();

            string combinedProperty = "";

            foreach (PropertyInfo propertyInfo in properties)
            {
                combinedProperty += "." + propertyInfo.Name;
            }

            return combinedProperty.Substring(1, combinedProperty.Length - 1);
        }

        /// <summary>
        /// Get a collection of <see cref="PropertyInfo"/> objects from a lambda expression representing the members to return.
        /// Example: if the lambda expression is item.SubItem.SubSubItem, this will return a <see cref="PropertyInfo"/> for SubItem, and another one for SubSubItem.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="expression"></param>
        /// <returns></returns>
        public static IEnumerable<PropertyInfo> GetProperties<T>(Expression<Func<T, object>> expression)
        {
            MemberExpression memberExpression = GetMemberExpression(expression);
            return GetPropertiesRecursive(memberExpression);
        }

        private static IEnumerable<PropertyInfo> GetPropertiesRecursive(MemberExpression memberExpression)
        {
            var propertyInfos = new List<PropertyInfo>();

            propertyInfos.Add((PropertyInfo)memberExpression.Member);

            if (memberExpression.Expression is MemberExpression)
            {
                propertyInfos.AddRange(GetPropertiesRecursive((MemberExpression)memberExpression.Expression));
            }

            return propertyInfos;
        }

        private static MemberExpression GetMemberExpression<T, TU>(Expression<Func<T, TU>> expression, bool enforceCheck = true)
        {
            MemberExpression memberExpression = null;
            if (expression.Body.NodeType == ExpressionType.Convert)
            {
                var body = (UnaryExpression)expression.Body;
                memberExpression = body.Operand as MemberExpression;
            }
            else if (expression.Body.NodeType == ExpressionType.MemberAccess)
            {
                memberExpression = expression.Body as MemberExpression;
            }

            if (enforceCheck && memberExpression == null)
            {
                throw new ArgumentException("Not a member access", "expression");
            }

            return memberExpression;
        }
    }

    /// <summary>
    /// Reflection exception base
    /// </summary>
    public class ReflectionBaseException : Exception
    {
        private readonly object @object;
        private readonly string propertyName;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="message"></param>
        /// <param name="propertyName"></param>
        /// <param name="object"></param>
        public ReflectionBaseException(string message, string propertyName, object @object) : base(message)
        {
            this.propertyName = propertyName;
            this.@object = @object;
        }

        /// <summary>
        /// Property Name which causes the problem
        /// </summary>
        public string PropertyName
        {
            get { return propertyName; }
        }

        /// <summary>
        /// Object we try to read
        /// </summary>
        public object Object
        {
            get { return @object; }
        }
    }

    /// <summary>
    /// Exception caused when the property does not exist on the object
    /// </summary>
    public class ReflectionPropertyDoesNotExistException : ReflectionBaseException
    {
        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="propertyName"></param>
        /// <param name="object"></param>
        public ReflectionPropertyDoesNotExistException(string propertyName, object @object)
            : base(
                "The property '" + propertyName + "' could not be found on the object '" + @object + "'.", propertyName,
                @object)
        {
        }
    }

    /// <summary>
    /// Exception caused when a property returns <see langword="null"/> and we need to access its child properties
    /// </summary>
    public class ReflectionPropertyParentIsNullException : ReflectionBaseException
    {
        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="propertyName"></param>
        /// <param name="object"></param>
        public ReflectionPropertyParentIsNullException(string propertyName, object @object)
            : base(
                "The property '" + propertyName + "' on the object '" + @object +
                "' returns null, so we can't access its children properties", propertyName, @object)
        {
        }
    }
}