#region Copyright
//       Copyright (c) 2009 - Antoine Jaussoin - http://www.jaussoin.com/blog/
// 
//        This program is free software: you can redistribute it and/or modify
//        it under the terms of the GNU General Public License as published by
//        the Free Software Foundation, either version 3 of the License, or
//        (at your option) any later version.
// 
//        This program is distributed in the hope that it will be useful,
//        but WITHOUT ANY WARRANTY; without even the implied warranty of
//        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//        GNU General Public License for more details.
// 
//        You should have received a copy of the GNU General Public License
//        along with this program.  If not, see <http://www.gnu.org/licenses/>.
#endregion
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using FT.Architecture.Example.Core.Entities.Cars;
using FT.Architecture.Example.Core.Entities.Humans;
using FT.Architecture.Example.Core.Entities.Pets;
using FT.Architecture.Example.Core.Entities.Subscriptions;
using FT.Architecture.Example.Core.Repositories;
using FT.Architecture.Repositories;
using FT.Architecture.UnitOfWork;

namespace FT.Architecture.Example.Core.Engines
{
    /// <summary>
    /// Database generator
    /// </summary>
    public class DatabaseGenerator
    {
        private readonly IUnitOfWork _unitOfWork;
        private readonly IRepository<IHuman> humanRepository;
        private readonly IRepository<IPet> petRepository;
        private readonly IGenericRepository<ICar, Guid> carRepository;
        private readonly IGenericRepository<Subscription, SubscriptionId> subscriptionRepository;

        public DatabaseGenerator(IUnitOfWork unitOfWork, IRepository<IHuman> humanRepository, IRepository<IPet> petRepository, IGenericRepository<ICar, Guid> carRepository, IGenericRepository<Subscription, SubscriptionId> subscriptionRepository)
        {
            _unitOfWork = unitOfWork;
            this.humanRepository = humanRepository;
            this.petRepository = petRepository;
            this.carRepository = carRepository;
            this.subscriptionRepository = subscriptionRepository;
        }

        /// <summary>
        /// Generate data in the database
        /// </summary>
        public void Generate(int numberOfEntities)
        {
            IList<IHuman> firstGeneration = GenerateHumans(numberOfEntities);

            MarryGeneration(firstGeneration);
            CreatePets(firstGeneration);

            IList<IHuman> currentGeneration = firstGeneration;

            //Generates three generations
            for (int i = 0; i < 7; i++)
            {
                currentGeneration =
                    ProduceGeneration(new List<IHuman>(from IHuman man in currentGeneration where man is Man select man));
            }
        }

        private IList<IHuman> ProduceGeneration(IList<IHuman> fathers)
        {
            IList<IHuman> nextGeneration = new List<IHuman>();

            foreach (IHuman man in fathers)
            {
                if (man.MarriedTo != null)
                {
                    for (int j = 0; j < RandomGenerator.GetRandomInt(2, 9); j++)
                    {
                        IHuman child = man.ProduceChild(
                            RandomGenerator.GetRandomName(),
                            RandomGenerator.GetRandomBirthDate(),
                            RandomGenerator.GetRandomPhoneNumber());

                        if (child is Woman)
                            ((Woman)child).LengthOfHairs = RandomGenerator.GetRandomInt(3, 50);

                        humanRepository.Save(child);
                        nextGeneration.Add(child);

                        GenerateSubscriptions(RandomGenerator.GetRandomInt(0, 9), child);
                        GenerateCars(RandomGenerator.GetRandomInt(1, 20), child);
                    }
                }
            }

            MarryGeneration(nextGeneration);
            CreatePets(nextGeneration);

            return nextGeneration;
        }

        private IList<IHuman> MarryGeneration(IList<IHuman> fullGeneration)
        {
            IList<IHuman> men = new List<IHuman>(from IHuman man in fullGeneration where man is Man select man);
            IList<IHuman> women = new List<IHuman>(from IHuman woman in fullGeneration where woman is Woman select woman);

            foreach (var man in men)
            {
                if (women.Count > 0)
                {
                    man.Mary(women[0]);
                    women.RemoveAt(0);
                }
            }

            return fullGeneration;
        }

        private IEnumerable<IPet> CreatePets(IList<IHuman> fullGeneration)
        {
            foreach (IHuman human in fullGeneration)
            {
                for (int j = 0; j < RandomGenerator.GetRandomInt(0, 2); j++)
                {
                    IPet pet = human.BuyAPet(RandomGenerator.GetRandomName());
                   petRepository.Save(pet);
                    //yield return pet;
                }
            }

            return null;
        }

        private IList<IHuman> GenerateHumans(int numberOfEntities)
        {
            IList<IHuman> humans = new List<IHuman>();

            for (int i = 0; i < numberOfEntities; i++)
            {
                IHuman human = HumanFactory.GetRandomHuman();
                human.Name = RandomGenerator.GetRandomName();
                human.Telephone = RandomGenerator.GetRandomPhoneNumber();
                human.DateBorn = RandomGenerator.GetRandomBirthDate();

                if (human is Woman)
                    ((Woman)human).LengthOfHairs = RandomGenerator.GetRandomInt(3, 50);

                humanRepository.Save(human);

                humans.Add(human);

                GenerateSubscriptions(RandomGenerator.GetRandomInt(0, 9), human);
                GenerateCars(RandomGenerator.GetRandomInt(1, 20), human);
            }


            return humans;
        }

        private void GenerateSubscriptions(int numberOfEntities, IHuman human)
        {
            for (int i = 0; i < numberOfEntities; i++)
            {
                Subscription subscription = new Subscription();
                subscription.Id.Human = human;
                subscription.Id.Type = (SubscriptionType)i;
                subscription.SubscribedOn = RandomGenerator.GetRandomDate(new DateTime(1990, 1, 1), DateTime.Today);
                subscription.FinishOn = subscription.SubscribedOn.AddMonths(RandomGenerator.GetRandomInt(1, 300));
                subscription.MonthlyPrice = (decimal)RandomGenerator.GetRandomInt(100, 100000) / 100;

                subscriptionRepository.Save(subscription);

                human.Subscriptions.Add(subscription);
            }
        }

        private void GenerateCars(int numberOfEntities, IHuman human)
        {
            for (int i = 0; i < numberOfEntities; i++)
            {
                ICar car = new BasicCar();
                car.Brand = RandomGenerator.GetRandomObject("Peugeot", "Renault", "Citron", "Vauxhall", "BMW",
                                                            "Mercedes", "Ferrari", "Opel", "Porsche", "Fiat",
                                                            "General Motors", "Ford", "Yamaha", "Suzuki");
                car.Colour = Color.FromKnownColor(RandomGenerator.GetRandomEnumeration<KnownColor>());
                car.Owner = human;
                car.Year = RandomGenerator.GetRandomInt(1990, 2010);

                carRepository.Save(car);

                human.Cars.Add(car);
            }
        }
    }
}