﻿#region Copyright
//       Copyright (c) 2009 - Antoine Jaussoin - http://www.jaussoin.com/blog/
// 
//        This program is free software: you can redistribute it and/or modify
//        it under the terms of the GNU General Public License as published by
//        the Free Software Foundation, either version 3 of the License, or
//        (at your option) any later version.
// 
//        This program is distributed in the hope that it will be useful,
//        but WITHOUT ANY WARRANTY; without even the implied warranty of
//        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//        GNU General Public License for more details.
// 
//        You should have received a copy of the GNU General Public License
//        along with this program.  If not, see <http://www.gnu.org/licenses/>.
#endregion

using System;
using FT.Architecture.Example.Core.Entities.Humans;
using FT.Architecture.Repositories;
using NHibernate;
using NUnit.Framework;

namespace FT.Architecture.Tests.Architecture
{
    [TestFixture]
    public class AttachAndDetachTest : BaseTest
    {
        [Test]
        public void TestAttachAndDetach()
        {
            Guid guid = Guid.NewGuid();
            const string name = "Germaine";
            const string name2 = "Gontrande";

            Man man;

            //Saving new man to DB
            unitOfWork.Start();
            man = new Man(null, null);
            man.Name = name;
            man.Telephone = guid.ToString();
            man.DateBorn = DateTime.Today.AddYears(-20);

            manRepository.Save(man);

            //Detaching the object
            manRepository.Detach(man);

            unitOfWork.Commit();

            unitOfWork.Start();
            man = manRepository.Attach(man, AttachMode.NoUpdate);
            man.Name = name2;
            unitOfWork.Commit();


            unitOfWork.Start();
            man = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.AreEqual(name2, man.Name, "The name hasn't been updated");
            unitOfWork.Commit();


            //We get rid of the created man
            unitOfWork.Start();
            Man man3 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            manRepository.Delete(man3);
            unitOfWork.Commit();

            //We check it doesn't exist anymore
            unitOfWork.Start();
            Man man4 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.IsNull(man4, "The man is not null");
            unitOfWork.Commit();

        }

        [Test]
        public void TestAttachAndDetachWithUpdate()
        {
            Guid guid = Guid.NewGuid();
            const string name = "Germaine";
            const string name2 = "Gontrande";

            Man man;

            //Saving new man to DB
            unitOfWork.Start();
            man = new Man(null, null);
            man.Name = name;
            man.Telephone = guid.ToString();
            man.DateBorn = DateTime.Today.AddYears(-20);

            manRepository.Save(man);

            //Detaching the object
            humanRepository.Detach(man);

            unitOfWork.Commit();


            //Changing the name OUTSIDE of a Unit Of Work context
            man.Name = name2;

            unitOfWork.Start();
            //Should not refresh the "man", but rather commit the changes
            //which occured outside of the UoW to the database (on the commit)
            man = manRepository.Attach(man, AttachMode.UpdateDatabase);
            Assert.AreEqual(name2, man.Name, "The name is back to " + man.Name);

            unitOfWork.Commit();


            unitOfWork.Start();
            //We load the man again, which should have its new name now
            man = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.AreEqual(name2, man.Name, "The name hasn't been updated");

            unitOfWork.Commit();


            //We get rid of the created man
            unitOfWork.Start();
            Man man3 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            manRepository.Delete(man3);

            unitOfWork.Commit();


            //We check it doesn't exist anymore
            unitOfWork.Start();
            Man man4 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.IsNull(man4, "The man is not null");

            unitOfWork.Commit();

        }

        [Test]
        public void TestingDetachingWithProxiesNoUpdate()
        {
            TestingDetachingWithProxies(AttachMode.NoUpdate);
        }

        [Test]
        public void TestingDetachingWithProxiesUpdateDatabase()
        {
            TestingDetachingWithProxies(AttachMode.UpdateDatabase);
        }

        [Test]
        public void TestingDetachingWithProxiesUpdateEntity()
        {
            TestingDetachingWithProxies(AttachMode.UpdateEntity);
        }

        [Test]
        public void When_The_Attach_Mode_Is_Not_A_Recognised_Enum_Value()
        {
            Assert.That(() => TestingDetachingWithProxies((AttachMode) int.MaxValue), Throws.Exception);
        }

        private void TestingDetachingWithProxies(AttachMode attachMode)
        {
            IHuman proxyHuman;
            unitOfWork.Start();
            IHuman human;
            do
            {
                human = humanRepository.GetRandom();
            }
            while (human.Father == null && human.Father.Father == null);

            //That should NOT be a proxy anymore
            proxyHuman = human.Father;

            Assert.IsFalse(proxyHuman.ToString().Contains("Proxy"));
            Assert.IsFalse(proxyHuman.Father.ToString().Contains("Proxy"));
            unitOfWork.Commit();

            unitOfWork.Start();
            //The object is not reattached, so it should throw an exception no problem
            string name = proxyHuman.Father.Name;
            Console.Out.WriteLine(name);

            //We reattach the proxy to the session
            proxyHuman = humanRepository.Attach(proxyHuman, attachMode);

            //This shouldn't throw an exception
            Assert.DoesNotThrow(() => Console.WriteLine(proxyHuman.Father.Name));
            unitOfWork.Commit();
        }

        [Test]
        public void TestingDetachingWithChanges()
        {
            IHuman originalHuman;
            string originalName;
            string modifiedName = Guid.NewGuid().ToString();

            //Get the original human
            unitOfWork.Start();
            originalHuman = humanRepository.GetRandom();
            originalName = originalHuman.Name;
            unitOfWork.Commit();

            unitOfWork.Start();
            IHuman modifiedHuman = humanRepository.GetById(originalHuman.Id);
            modifiedHuman.Name = modifiedName;
            unitOfWork.Commit();


            //Name on the original human shouldn't have changed
            Assert.AreEqual(originalName, originalHuman.Name);

            unitOfWork.Start();
            humanRepository.Attach(originalHuman, AttachMode.NoUpdate);
            //Name on the original human shouldn't have changed, because we don't refresh
            Assert.AreEqual(originalName, originalHuman.Name);
            unitOfWork.Commit();

            unitOfWork.Start();
            humanRepository.Attach(originalHuman, AttachMode.UpdateEntity);
            //Name on the original human should have updated to the new name
            Assert.AreEqual(modifiedName, originalHuman.Name);
            //We revert it back so it's not committed for ever on the database
            originalHuman.Name = originalName;
            unitOfWork.Commit();


            unitOfWork.Start();
            humanRepository.Attach(originalHuman, AttachMode.UpdateEntity);
            //Name on the original human should be back to what it was
            Assert.AreEqual(originalName, originalHuman.Name);
            unitOfWork.Commit();
        }
    }
}