﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using FT.Architecture.Example.Core.Entities.Cars;
using FT.Architecture.Example.Core.Entities.Humans;
using FT.Architecture.Example.Core.Entities.Pets;
using FT.Architecture.NHibernate.Repositories;
using FT.Architecture.NHibernate.SessionManagement;
using FT.Architecture.Repositories;
using NHibernate;
using NUnit.Framework;
using Rhino.Mocks;

namespace FT.Architecture.Tests.Architecture
{
    public class DeleteTests : BaseTest
    {
        [TestFixture]
        public class When_Deleting_An_Entity_And_Its_Collections_By_Cascade : DeleteTests
        {
            [Test]
            public void It_Shouldnt_Fail_When_Resetting_The_Id()
            {
                // Arrange
                unitOfWork.Start();
                IHuman human = new Man();
                human.Name = "Tartanpion";
                human.Telephone = "";
                human.DateBorn = DateTime.Today;
                humanRepository.Save(human);
                IPet pet1 = new Dog { Name = "Médor", BorrowedBy = human};
                IPet pet2 = new Dog { Name = "Pétrus", BorrowedBy = human };
                human.BorrowedPets.Add(pet1);
                human.BorrowedPets.Add(pet2);
                humanRepository.Save(human);
                unitOfWork.Flush();

                // Act
                humanRepository.Delete(human);

                // Assert
                Assert.That(human.Id, Is.EqualTo(0));
                unitOfWork.Commit();
            }

            [Test]
            public void It_Shouldnt_Fail_When_NOT_Resetting_The_Id()
            {
                // Arrange
                unitOfWork.Start();
                IHuman human = new Man();
                human.Name = "Tartanpion";
                human.Telephone = "";
                human.DateBorn = DateTime.Today;
                humanRepository.Save(human);
                IPet pet1 = new Dog { Name = "Médor", BorrowedBy = human };
                IPet pet2 = new Dog { Name = "Pétrus", BorrowedBy = human };
                human.BorrowedPets.Add(pet1);
                human.BorrowedPets.Add(pet2);
                humanRepository.Save(human);
                unitOfWork.Flush();

                // Act
                humanRepository.Delete(human, false);

                // Assert
                Assert.That(human.Id, Is.Not.EqualTo(0));
                unitOfWork.Commit();
            }

            [Test]
            public void It_Should_Work_When_Deletting_A_Batch_Of_Objects()
            {
                // Arrange
                unitOfWork.Start();
                var humans = subscriptionRepository.GetAll();

                // Act
                subscriptionRepository.Delete(humans);

                // Assert
                foreach (var human in humans)
                {
                    Assert.That(human.Id, Is.Null);
                }
                unitOfWork.Rollback();
            }
        }

        [TestFixture]
        public class When_Deleting_A_Batch_Of_Entities : DeleteTests
        {
            [Test]
            public void The_Session_Should_Be_Flushed_Once_Only()
            {
                // Arrange
                var mockSession = MockRepository.GenerateMock<ISession>();
                mockSession.Stub(x => x.Flush());
                var mockSessionProvider = MockRepository.GenerateMock<ISessionProviderService>();
                mockSessionProvider.Stub(x => x.GetSession()).Return(mockSession);
                var systemUnderTest = new NHibernateGenericRepository<ICar, Guid>(mockSessionProvider, null, logService);
                var allCars = GetCars();

                // Act
                systemUnderTest.Delete(allCars);

                // Assert
                mockSession.AssertWasCalled(x => x.Flush(), x => x.Repeat.Once());
            }

            [Test]
            public void All_Ids_Should_Be_Reset_When_Asking_For_Reset_Ids()
            {
                // Arrange
                var mockSession = MockRepository.GenerateMock<ISession>();
                mockSession.Stub(x => x.Flush());
                var mockSessionProvider = MockRepository.GenerateMock<ISessionProviderService>();
                mockSessionProvider.Stub(x => x.GetSession()).Return(mockSession);
                var systemUnderTest = new NHibernateGenericRepository<ICar, Guid>(mockSessionProvider, null, logService);
                var allCars = GetCars();

                // Act
                systemUnderTest.Delete(allCars);

                // Assert
                foreach (var car in allCars)
                {
                    Assert.That(car.Id, Is.EqualTo(default(Guid)));
                }
            }

            [Test]
            public void All_Ids_Should_Be_Reset_When_Asking_For_Not_Reset_Ids()
            {
                // Arrange
                var mockSession = MockRepository.GenerateMock<ISession>();
                mockSession.Stub(x => x.Flush());
                var mockSessionProvider = MockRepository.GenerateMock<ISessionProviderService>();
                mockSessionProvider.Stub(x => x.GetSession()).Return(mockSession);
                var systemUnderTest = new NHibernateGenericRepository<ICar, Guid>(mockSessionProvider, null, logService);
                var allCars = GetCars();

                // Act
                systemUnderTest.Delete(allCars, false);

                // Assert
                foreach (var car in allCars)
                {
                    Assert.That(car.Id, Is.Not.EqualTo(default(Guid)));
                }
            }

            private IList<ICar> GetCars()
            {
                return new List<ICar>
                {
                    new BasicCar { Id = Guid.NewGuid() } ,
                    new BasicCar { Id = Guid.NewGuid() } ,
                    new BasicCar { Id = Guid.NewGuid() } ,
                    new BasicCar { Id = Guid.NewGuid() }
                };
            }
        }
    }
}
