﻿#region Copyright
//       Copyright (c) 2009 - Antoine Jaussoin - http://www.jaussoin.com/blog/
// 
//        This program is free software: you can redistribute it and/or modify
//        it under the terms of the GNU General Public License as published by
//        the Free Software Foundation, either version 3 of the License, or
//        (at your option) any later version.
// 
//        This program is distributed in the hope that it will be useful,
//        but WITHOUT ANY WARRANTY; without even the implied warranty of
//        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//        GNU General Public License for more details.
// 
//        You should have received a copy of the GNU General Public License
//        along with this program.  If not, see <http://www.gnu.org/licenses/>.
#endregion

using System.Collections.Generic;
using System.Linq;
using FT.Architecture.Example.Core.Entities;
using FT.Architecture.Example.Core.Entities.Humans;
using FT.Architecture.Example.Core.Entities.Pets;
using FT.Architecture.Example.Core.Entities.Subscriptions;
using FT.Architecture.Utils.ExtensionMethods;
using NUnit.Framework;

namespace FT.Architecture.Tests.Architecture
{
    [TestFixture]
    public class GenericRepositoryTest : BaseTest
    {
        [Test]
        public void TestingGetById()
        {
            unitOfWork.Start();
            //Get a random dog to get a valid Id
            Dog randomDog = dogRepository.GetRandom();
            long dogId = randomDog.Id;

            //Get this dog back from its id
            Dog dogById = dogRepository.GetById(dogId);

            Assert.AreEqual(randomDog, dogById, "The dogs are different");
            unitOfWork.Commit();
        }

        [Test]
        public void TestingGetByMemberSimple()
        {
            unitOfWork.Start();
            //Get a random dog to get a valid Id
            IList<Dog> dogs = dogRepository.GetAllByMember(d => d.Name, "Fra%");

            //Check if all the returned dog have a name starting by Fra
            foreach (Dog dog in dogs)
            {
                Assert.IsTrue(dog.Name.StartsWith("Fra"), "The dog's name doesn't start by Fra: " + dog.Name);
            }
            unitOfWork.Commit();
        }

        [Test]
        public void TestingGetByMemberComplex()
        {
            unitOfWork.Start();
            //Get a random dog to get a valid Id
            IList<Woman> women = womanRepository.GetAllByMember(d => d.Father.Father.Father.Name, "Fra%");

            //Check if all the returned women grand-grand-father's have a name starting by Fra
            foreach (Woman woman in women)
            {
                Assert.IsTrue(woman.Father.Father.Father.Name.StartsWith("Fra"),
                              "The woman grand-grand-father's name doesn't start by Fra: " +
                              woman.Father.Father.Father.Name);
            }
            unitOfWork.Commit();
        }

        [Test]
        public void TestingFetchingSimple()
        {
            Woman woman;
            unitOfWork.Start();
            do
            {
                //Get a random dog to get a valid Id
                Woman randomDog = womanRepository.GetRandom();
                long dogId = randomDog.Id;

                //Get a random dog to get a valid Id
                woman = womanRepository.GetById(dogId,
                                                    d => d.Father,
                                                    d => d.Mother);

                womanRepository.Detach(woman);

            } while (woman.Father == null || woman.Mother == null);

            unitOfWork.Commit();

            //The woman is now detached from its Repository, but should be able to read Father andd Mother values
            Assert.IsNotNull(woman.Father, "Woman's father can't be null");
            Assert.IsNotNull(woman.Mother, "Woman's mother can't be null");

            Assert.IsNotNull(woman.Father.Name, "Woman's father can't be null");
            Assert.IsNotNull(woman.Mother.Name, "Woman's mother can't be null");

            try
            {
                //That should break
                string test = woman.Father.Father.Name;
                Assert.Fail("The previous line of code should have break");
            }
            catch
            {

            }

        }

        [Test]
        public void TestingFetchingComplex()
        {
            Woman woman;
            unitOfWork.Start();
            do
            {
                //Get a random dog to get a valid Id
                Woman randomDog = womanRepository.GetRandom();
                long dogId = randomDog.Id;

                //Get a random dog to get a valid Id
                woman = womanRepository.GetById(dogId,
                                                    d => d.Father.Father,
                                                    d => d.Mother.Mother);

                womanRepository.Detach(woman);

            } while (woman.Father.Father == null || woman.Mother.Mother == null);

            unitOfWork.Commit();

            //The woman is now detached from its Repository, but should be able to read Father andd Mother values
            Assert.IsNotNull(woman.Father.Father, "Woman's father can't be null");
            Assert.IsNotNull(woman.Mother.Mother, "Woman's mother can't be null");

            Assert.IsNotNull(woman.Father.Father.Name, "Woman's father can't be null");
            Assert.IsNotNull(woman.Mother.Mother.Name, "Woman's mother can't be null");

            try
            {
                //That should break
                string test = woman.Father.Father.Father.Name;
                Assert.Fail("The previous line of code should have break");
            }
            catch
            {

            }

        }

        [Test]
        public void TestingCount()
        {
            unitOfWork.Start();
            IList<Dog> allTheDogs = dogRepository.GetAll();
            long count = dogRepository.GetCount();

            Assert.AreEqual(allTheDogs.Count, count, "The count method is not working");
            unitOfWork.Commit();
        }

        [Test]
        public void TestingRandom()
        {
            unitOfWork.Start();
            Dog randomDog = dogRepository.GetRandom();

            Assert.IsNotNull(randomDog, "The random method returned a null entity, which should never happen, unless the table is empty");
            unitOfWork.Commit();
        }

        [Test]
        public void TestingCompositeId()
        {
            unitOfWork.Start();
            //First we get a random subscription
            Subscription subscription = subscriptionRepository
                .GetRandom();

            //Then we figure out to what it is attached to
            IHuman human = subscription.Human;
            SubscriptionType type = subscription.Type;

            //Then we try to load it back from its ID
            Subscription subscriptionById = subscriptionRepository
                .GetById(new SubscriptionId(human, type));

            //We check whether they are equal (should be!!)
            Assert.AreEqual(subscription, subscriptionById);

            //We try to get all the subscription of this given human
            var allSubscriptions = subscriptionRepository
                .GetAllByMember(x => x.Id.Human, human);

            //This shouldn't be empty, as the original subscription should be part of it at least
            Assert.IsNotEmpty(allSubscriptions.ToArray());
            unitOfWork.Commit();
        }

        [Test]
        public void TestingUpperLevelInterface()
        {
            unitOfWork.Start();
            var allEntitiesWhoCanSpeak = canSpeakRepository.GetCount();
            var allHumans = humanRepository.GetCount();
            var allPets = petRepository.GetCount();

            Assert.AreEqual(allHumans + allPets, allEntitiesWhoCanSpeak, "The number if ICanSpeak should be equal to the sum of all IHuman and all IPet");
            unitOfWork.Commit();
        }

        [Test, Ignore]
        public void TestingUpperLevelInterfaceLinq()
        {
            unitOfWork.Start();
            var allEntitiesWhoCanSpeak = canSpeakRepository.GetLinq();

            var allEntitiesWhoCanSpeakStartingMultipleOf33 = (
                from ICanSpeak entity in allEntitiesWhoCanSpeak.ToArray()
                where entity.Id % 33 == 0
                select entity).ToArray();

            //Checking we have results
            Assert.IsNotEmpty(allEntitiesWhoCanSpeakStartingMultipleOf33);

            foreach (var entity in allEntitiesWhoCanSpeakStartingMultipleOf33)
            {
                //Check if every returned entities is actually greater than last id - 1000
                Assert.IsTrue(entity.Id % 33 == 0);
            }

            //Checking we have results of all kind of entities
            var humanEntities = from ICanSpeak entity in allEntitiesWhoCanSpeakStartingMultipleOf33
                                where entity.GetType().Is(typeof(IHuman))
                                select entity;

            var petEntities = from ICanSpeak entity in allEntitiesWhoCanSpeakStartingMultipleOf33
                              where entity.GetType().Is(typeof(IPet))
                              select entity;

            Assert.IsNotEmpty(humanEntities.ToArray());
            Assert.IsNotEmpty(petEntities.ToArray());

            //Checking that the sum of the two sets is equal to the original set
            Assert.AreEqual(humanEntities.Count() + petEntities.Count(), allEntitiesWhoCanSpeakStartingMultipleOf33.Count());
            unitOfWork.Commit();
        }

        [Test]
        public void TestUnProxy()
        {
            unitOfWork.Start();
            IHuman human;

            do
            {
                human = humanRepository.GetRandom();
            } while (human.Father == null);

            var proxy = human.Father;
            var nonProxy = humanRepository.UnProxy(proxy);

            //One is a proxy, the other one is not
            Assert.AreEqual(proxy, nonProxy); //Equals() should say yes
            unitOfWork.Commit();
        }

        [Test]
        public void TestGetNextId()
        {
            unitOfWork.Start();
            long nextId = humanRepository.GetNextId();
            IHuman humanWhichShouldNotExist = humanRepository.GetById(nextId);

            Assert.IsNull(humanWhichShouldNotExist);
            unitOfWork.Commit();
        }

        [Test]
        public void TestGetConcreteType()
        {
            unitOfWork.Start();
            IHuman human;

            do
            {
                human = humanRepository.GetRandom();
            } while (human.Father == null);

            var proxy = human.Father;
            var nonProxy = humanRepository.UnProxy(proxy);

            //Test the GetType returns the same as GetConcreteType()
            Assert.AreEqual(proxy.GetType(), humanRepository.GetConcreteType(proxy));
            Assert.AreEqual(nonProxy.GetType(), humanRepository.GetConcreteType(proxy));
            unitOfWork.Commit();
        }

        [Test]
        public void TestGetAllWithFetch()
        {
            unitOfWork.Start();
            var allSupermen = supermanRepository.GetAll();
            var allSupermenWithFetch = supermanRepository.GetAll(x => x.Father);
            var allSupermenWithSuperfetch = supermanRepository.GetAll(x => x.Father.Father);

            Assert.AreEqual(allSupermen.Count, allSupermenWithFetch.Count);
            Assert.AreEqual(allSupermen.Count, allSupermenWithSuperfetch.Count);
            unitOfWork.Commit();
        }

        [Test]
        public void TestGetAllWithFetchAndSort()
        {
            unitOfWork.Start();
            var allSupermen = supermanRepository.GetAll();
            var allSupermenWithFetchAndSort = supermanRepository.GetAll(x => x.Name, true, x => x.Father);
            var allSupermenWithSuperfetchAndSort = supermanRepository.GetAll(x => x.Name, true, x => x.Father.Father);

            Assert.AreEqual(allSupermen.Count, allSupermenWithFetchAndSort.Count);
            Assert.AreEqual(allSupermen.Count, allSupermenWithSuperfetchAndSort.Count);
            unitOfWork.Commit();
        }

        [Test]
        public void TestGetAllByMultipleMembers()
        {
            unitOfWork.Start();
            var allStartingWithAnt = humanRepository.GetAllByMember(x => x.Name, "Ant%", x => x.CV, "%", x => x.Father);
            var allStartingWithAnt2 = humanRepository.GetAllByMember(x => x.Name, "Ant%", x => x.CV, "%", x => x.Telephone, "%", x => x.Father);

            Assert.AreEqual(allStartingWithAnt.Count, allStartingWithAnt2.Count);

            unitOfWork.Commit();
        }


    }
}
