﻿using System;
using System.Linq;
using FT.Architecture.Example.Core.Entities.Humans;
using FT.Architecture.Example.Data.Repositories;
using FT.Architecture.NHibernate.Queries;
using FT.Architecture.Queries;
using NUnit.Framework;

namespace FT.Architecture.Tests.Architecture
{
    public abstract class QueryTests : BaseTest
    {
        private NHibernateHumanRepository repository;
        private IQueryExecutorService systemUnderTest;

        public override void SetUp()
        {
            base.SetUp();
            unitOfWork.Start();
        }

        [TearDown]
        public void TearDown()
        {
            unitOfWork.Rollback();
        }

        [Test]
        public void A_Simple_Select_Should_Return_All_Entities()
        {
            // Arrange
            var referenceResults = repository.GetAll();

            // Act
            var query = repository.CreateQuery();
            var results = repository.FindBy(query);

            // Assert
            Assert.That(referenceResults.Count, Is.EqualTo(results.Count()));
        }

        [Test]
        public void Testing_A_Single_Where_Clause_Like()
        {
            // Arrange
            var referenceResults = repository.GetAllByMember(x => x.Name, "%A%");

            // Act
            var query = repository.CreateQuery().Where(x => x.Name, CriteriaOperator.Like, "%A%");
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.Count(), Is.EqualTo(referenceResults.Count));
        }

        [Test]
        public void Testing_A_Single_Where_Clause_Equal()
        {
            // Arrange
            var referenceResults = repository.GetAllByMember(x => x.Name, "Antoine");

            // Act
            var query = repository.CreateQuery().Where(x => x.Name, CriteriaOperator.Equal, "Antoine");
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.Count(), Is.EqualTo(referenceResults.Count));
        }

        [Test]
        public void Testing_A_Single_Where_Clause_Different()
        {
            // Arrange
            var session = repository.SessionProvider.GetSession();
            var referenceResults = session.CreateQuery("from IHuman e where e.Name <> :name")
                .SetString("name", "Antoine")
                .List<IHuman>();

            // Act
            var query = repository.CreateQuery().Where(x => x.Name, CriteriaOperator.NotEqual, "Antoine");
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.Count(), Is.EqualTo(referenceResults.Count));
        }

        [Test]
        public void Testing_A_Single_Where_Clause_GreaterThan()
        {
            // Arrange
            var randomHuman = repository.GetRandom();
            var session = repository.SessionProvider.GetSession();
            var referenceResults = session.CreateQuery("from IHuman e where e.Id > :id")
                .SetInt64("id", randomHuman.Id)
                .List<IHuman>();

            // Act
            var query = repository.CreateQuery().Where(x => x.Id, CriteriaOperator.GreaterThan, randomHuman.Id);
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.Count(), Is.EqualTo(referenceResults.Count));
        }

        [Test]
        public void Testing_A_Single_Where_Clause_LowerThan()
        {
            // Arrange
            var randomHuman = repository.GetRandom();
            var session = repository.SessionProvider.GetSession();
            var referenceResults = session.CreateQuery("from IHuman e where e.Id < :id")
                .SetInt64("id", randomHuman.Id)
                .List<IHuman>();

            // Act
            var query = repository.CreateQuery().Where(x => x.Id, CriteriaOperator.LessThan, randomHuman.Id);
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.Count(), Is.EqualTo(referenceResults.Count));
        }

        [Test]
        public void Testing_A_Single_Where_Clause_GreaterThanOrEqual()
        {
            // Arrange
            var randomHuman = repository.GetRandom();
            var session = repository.SessionProvider.GetSession();
            var referenceResults = session.CreateQuery("from IHuman e where e.Id >= :id")
                .SetInt64("id", randomHuman.Id)
                .List<IHuman>();

            // Act
            var query = repository.CreateQuery().Where(x => x.Id, CriteriaOperator.GreaterThanOrEqual, randomHuman.Id);
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.Count(), Is.EqualTo(referenceResults.Count));
        }

        [Test]
        public void Testing_A_Single_Where_Clause_LowerThanOrEqual()
        {
            // Arrange
            var randomHuman = repository.GetRandom();
            var session = repository.SessionProvider.GetSession();
            var referenceResults = session.CreateQuery("from IHuman e where e.Id <= :id")
                .SetInt64("id", randomHuman.Id)
                .List<IHuman>();

            // Act
            var query = repository.CreateQuery().Where(x => x.Id, CriteriaOperator.LessThanOrEqual, randomHuman.Id);
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.Count(), Is.EqualTo(referenceResults.Count));
        }

        [Test]
        public void Testing_Two_Where_Clause_With_And()
        {
            // Arrange
            var referenceResults = repository.GetAllByMember(x => x.Name, "%A%", x => x.Telephone, "%8%");

            // Act
            var query = repository.CreateQuery()
                .SetOperator(QueryOperator.And)
                .Where(x => x.Name, CriteriaOperator.Like, "%A%")
                .Where(x => x.Telephone, CriteriaOperator.Like, "%8%");
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.Count(), Is.EqualTo(referenceResults.Count));
        }

        [Test]
        public void Testing_Two_Where_Clause_With_Or()
        {
            // Arrange
            var session = repository.SessionProvider.GetSession();
            var referenceResults = session.CreateQuery("from IHuman e where e.Name like :name or e.Telephone like :telephone")
                .SetString("name", "%A%")
                .SetString("telephone", "%8%")
                .List<IHuman>();

            // Act
            var query = repository.CreateQuery()
                .SetOperator(QueryOperator.Or)
                .Where(x => x.Name, CriteriaOperator.Like, "%A%")
                .Where(x => x.Telephone, CriteriaOperator.Like, "%8%");
            var results = repository.FindBy(query);

            // Assert
            Assert.That(referenceResults.Count, Is.EqualTo(results.Count()));
        }

        [Test]
        public void Testing_Pagination()
        {
            // Arrange
            var referenceResults = repository.GetAll();

            // Act
            var query = repository.CreateQuery()
                .Paginate(1, 20);
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.Count(), Is.EqualTo(20));
            Assert.That(results.First().Id, Is.EqualTo(referenceResults[1].Id));
        }

        [Test]
        public void Testing_Ordering()
        {
            // Arrange
            var referenceResults = repository.GetAll(x => x.Name, true);

            // Act
            var query = repository.CreateQuery();
            query.Sort.By(x => x.Name).Ascending();
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.Count(), Is.EqualTo(referenceResults.Count));
            Assert.That(results.First().Id, Is.EqualTo(referenceResults[0].Id));
            Assert.That(results.Skip(1).First().Id, Is.EqualTo(referenceResults[1].Id));
            Assert.That(results.Skip(2).First().Id, Is.EqualTo(referenceResults[2].Id));
            Assert.That(results.Last().Id, Is.EqualTo(referenceResults.Last().Id));
        }

        [Test]
        public void Testing_Ordering_With_Two_Criteria()
        {
            // Arrange
            var referenceResults = repository.GetHumansOrderedByNameThenTelephone();

            // Act
            var query = repository.CreateQuery();
            query.Sort.By(x => x.Name).Ascending()
                      .And(x => x.Telephone).Descending();
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.Count(), Is.EqualTo(referenceResults.Count));
            Assert.That(results.First().Id, Is.EqualTo(referenceResults[0].Id));
            Assert.That(results.Skip(1).First().Id, Is.EqualTo(referenceResults[1].Id));
            Assert.That(results.Skip(2).First().Id, Is.EqualTo(referenceResults[2].Id));
            Assert.That(results.Last().Id, Is.EqualTo(referenceResults.Last().Id));
        }

        [Test]
        public void Testing_Ordering_With_Where_Clause()
        {
            // Arrange
            var referenceResults = repository.GetHumansOrderedByNameThenTelephone();

            // Act
            var query = repository.CreateQuery()
                .Sort.By(x => x.Name).Ascending()
                .And(x => x.Telephone).Descending()
                .Done()
                .Where(x => x.Name, CriteriaOperator.Like, "%");
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.Count(), Is.EqualTo(referenceResults.Count));
            Assert.That(results.First().Id, Is.EqualTo(referenceResults[0].Id));
            Assert.That(results.Skip(1).First().Id, Is.EqualTo(referenceResults[1].Id));
            Assert.That(results.Skip(2).First().Id, Is.EqualTo(referenceResults[2].Id));
            Assert.That(results.Last().Id, Is.EqualTo(referenceResults.Last().Id));
        }

        [Test]
        public void Testing_With_Alias_Properties()
        {
            // Arrange
            var session = repository.SessionProvider.GetSession();
            var referenceResults = session.CreateQuery("from IHuman e where e.Mother.Name like :motherName")
                .SetString("motherName", "%A%")
                .List<IHuman>();

            // Act
            var query = repository.CreateQuery();
            query.Where(x => x.Mother.Name, CriteriaOperator.Like, "%A%");
            var results = repository.FindBy(query);

            // Assert
            Assert.That(referenceResults.Count, Is.EqualTo(results.Count()));
        }

        [Test]
        public void Testing_Sub_Queries()
        {
            // Arrange
            var session = repository.SessionProvider.GetSession();
            var referenceResults = session.CreateQuery("from IHuman e where e.Name like :name and e.Telephone like :telephone")
                .SetString("name", "%A%")
                .SetString("telephone", "%8%")
                .List<IHuman>();

            // Act
            var query = repository.CreateQuery();
            query.Where(x => x.Name, CriteriaOperator.Like, "%A%");
            var subQuery = repository.CreateQuery();
            subQuery.Where(x => x.Telephone, CriteriaOperator.Like, "%8%");
            query.AddSubQuery(subQuery);
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.Count(), Is.EqualTo(referenceResults.Count));
            Assert.That(results.First().Id, Is.EqualTo(referenceResults[0].Id));
        }

        [Test]
        public void Testing_Sub_Queries_With_Different_Operators()
        {
            // Arrange
            var session = repository.SessionProvider.GetSession();
            var referenceResults = session.CreateQuery("from IHuman e where (e.Name like :p1 or e.Name like :p2 or (e.Telephone like :p3 and e.Telephone like :p4))")
                .SetString("p1", "%A%")
                .SetString("p2", "%B%")
                .SetString("p3", "%7%")
                .SetString("p4", "%8%")
                .List<IHuman>();

            var query = repository.CreateQuery()
                .SetOperator(QueryOperator.Or)
                .Where(x => x.Name, CriteriaOperator.Like, "%A%")
                .Where(x => x.Name, CriteriaOperator.Like, "%B%");

            var subQuery = repository.CreateQuery()
                .SetOperator(QueryOperator.And)
                .Where(x => x.Telephone, CriteriaOperator.Like, "%7%")
                .Where(x => x.Telephone, CriteriaOperator.Like, "%8%");

            query.AddSubQuery(subQuery);

            // Act
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.Count(), Is.EqualTo(referenceResults.Count));
            Assert.That(results.First().Id, Is.EqualTo(referenceResults[0].Id));
        }

        [Test]
        public void Testing_Fetching()
        {
            // Arrange
            var query = repository.CreateQuery();
            query.Where(x => x.Father, CriteriaOperator.IsNotNull);
            query.Fetch(x => x.Father);

            // Act
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.First().Father.GetType(), Is.EqualTo(((IHuman)results.First().Father).GetTypeUnproxied()));
        }

        [Test]
        public void Testing_When_Sub_Query_Is_Empty()
        {
            // Arrange
            var query = repository.CreateQuery();
            query.Where(x => x.Father, CriteriaOperator.IsNotNull);
            var subQuery = repository.CreateQuery();
            query.AddSubQuery(subQuery);

            // Act
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results, Is.Not.Null);
        }

        [Test]
        public void Testing_Is_Null()
        {
            // Arrange
            var session = repository.SessionProvider.GetSession();
            var referenceResults = session.CreateQuery("from IHuman e where e.Father is null")
                .List<IHuman>();

            // Act
            var query = repository.CreateQuery();
            query.Where(x => x.Father, CriteriaOperator.IsNull);
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.Count(), Is.EqualTo(referenceResults.Count));
        }

        [Test]
        public void Testing_Is_Not_Null()
        {
            // Arrange
            var session = repository.SessionProvider.GetSession();
            var referenceResults = session.CreateQuery("from IHuman e where e.Father is not null")
                .List<IHuman>();

            // Act
            var query = repository.CreateQuery();
            query.Where(x => x.Father, CriteriaOperator.IsNotNull);
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.Count(), Is.EqualTo(referenceResults.Count));
        }

        [Test]
        public void Testing_Sound_Like()
        {
            // Arrange
            var session = repository.SessionProvider.GetSession();
            Man man = new Man();
            man.Name = "AnneClaireMarie";
            man.DateBorn = new DateTime(1982, 11, 3);
            session.Save(man);
            session.Flush();

            // Act
            var query = repository.CreateQuery();
            query.Where(x => x.Name, CriteriaOperator.SoundsLike, "AnneClaireMary");
            var results = repository.FindBy(query);

            // Assert
            Assert.That((from result in results where result.Name == "AnneClaireMarie" select result).Count(), Is.EqualTo(1));
        }

        [Test]
        public void Testing_Fetching_In_Sub_Query()
        {
            // Arrange
            var query = repository.CreateQuery();
            query.Where(x => x.Father, CriteriaOperator.IsNotNull);
            var subQuery = repository.CreateQuery();
            subQuery.Fetch(x => x.Father);
            query.AddSubQuery(subQuery);

            // Act
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.First().Father.GetType(), Is.EqualTo(((IHuman)results.First().Father).GetTypeUnproxied()));
        }

        [Test]
        public void Testing_Nested_Or_And_Or()
        {
            // Arrange
            var query = repository.CreateQuery();
            query.SetOperator(QueryOperator.Or);

            var subQuery11 = repository.CreateQuery();
            subQuery11.SetOperator(QueryOperator.Or);
            subQuery11.Where(x => x.Name, CriteriaOperator.Like, "%A%");
            subQuery11.Where(x => x.Telephone, CriteriaOperator.Like, "%A%");

            var subQuery12 = repository.CreateQuery();
            subQuery12.SetOperator(QueryOperator.Or);
            subQuery12.Where(x => x.Name, CriteriaOperator.Like, "%1%");
            subQuery12.Where(x => x.Telephone, CriteriaOperator.Like, "%1%");

            var subQuery1 = repository.CreateQuery();
            subQuery1.SetOperator(QueryOperator.And);
            subQuery1.AddSubQuery(subQuery11);
            subQuery1.AddSubQuery(subQuery12);

            var subQuery21 = repository.CreateQuery();
            subQuery21.SetOperator(QueryOperator.Or);
            subQuery21.Where(x => x.Name, CriteriaOperator.Like, "%A 1%");
            subQuery21.Where(x => x.Telephone, CriteriaOperator.Like, "%A 1%");

            var subQuery2 = repository.CreateQuery();
            subQuery2.SetOperator(QueryOperator.And);
            subQuery2.AddSubQuery(subQuery21);

            query.AddSubQuery(subQuery1);
            query.AddSubQuery(subQuery2);

            var session = repository.SessionProvider.GetSession();
            var referenceResults = session.CreateQuery(@"
from IHuman e where 

    ((e.Name like '%A%' or e.Telephone like '%A%') and
     (e.Name like '%1%' or e.Telephone like '%1%'))
    or
    (e.Name like '%A 1%' or e.Telephone like '%A 1%')
")
                .List<IHuman>();

            // Act
            var results = repository.FindBy(query);

            // Assert
            Assert.That(results.Count(), Is.EqualTo(referenceResults.Count()));
        }

        [TestFixture]
        public class When_Querying_Using_Hql : QueryTests
        {
            public override void SetUp()
            {
                base.SetUp();

                systemUnderTest = new HqlExecutorService(sessionProviderService);
                repository = new NHibernateHumanRepository(sessionProviderService, systemUnderTest, logService);
            }
        }
    }
}
