﻿#region Copyright
//       Copyright (c) 2009 - Antoine Jaussoin - http://www.jaussoin.com/blog/
// 
//        This program is free software: you can redistribute it and/or modify
//        it under the terms of the GNU General Public License as published by
//        the Free Software Foundation, either version 3 of the License, or
//        (at your option) any later version.
// 
//        This program is distributed in the hope that it will be useful,
//        but WITHOUT ANY WARRANTY; without even the implied warranty of
//        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//        GNU General Public License for more details.
// 
//        You should have received a copy of the GNU General Public License
//        along with this program.  If not, see <http://www.gnu.org/licenses/>.
#endregion

using System;
using System.Collections.Generic;
using System.Threading;
using FT.Architecture.Example.Core.Entities.Humans;
using FT.Architecture.Example.Core.Entities.Pets;
using NUnit.Framework;

namespace FT.Architecture.Tests.Architecture
{
    [TestFixture]
    public class StatelessUnitOfWorkTest : BaseTest
    {
        [Test]
        public void TestInNormalConditions()
        {
            Guid guid = Guid.NewGuid();

            var man = new Man(null, null);
            man.Name = "Bob";
            man.Telephone = guid.ToString();
            man.DateBorn = DateTime.Today.AddYears(-20);

            //Saving new man to DB using a stateless session
            statelessUnitOfWork.Start();
            manRepository.BulkSave(new List<Man> {man});
            statelessUnitOfWork.Commit();

            //Get the man from the DB using his fake (guid) phone number
            unitOfWork.Start();
            Man man2 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.AreEqual("Bob", man2.Name);

            unitOfWork.Commit();

            //We get rid of the created man
            unitOfWork.Start();
            Man man3 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            manRepository.Delete(man3);

            unitOfWork.Commit();

            //We check it doesn't exist anymore
            unitOfWork.Start();
            Man man4 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.IsNull(man4, "The man is not null");

            unitOfWork.Commit();
        }

        [Test]
        public void TestIsStartedWhenStarted()
        {
            statelessUnitOfWork.Start();
            Assert.That(statelessUnitOfWork.IsStarted(), Is.True);
            statelessUnitOfWork.Commit();
        }

        [Test]
        public void TestIsStartedWhenCommitted()
        {
            statelessUnitOfWork.Start();
            statelessUnitOfWork.Commit();
            Assert.That(statelessUnitOfWork.IsStarted(), Is.Not.True);
        }

        [Test]
        public void TestIsStartedWhenRollback()
        {
            statelessUnitOfWork.Start();
            statelessUnitOfWork.Rollback();
            Assert.That(statelessUnitOfWork.IsStarted(), Is.Not.True);
        }

        [Test]
        public void TestIsStartedWhenNotStarted()
        {
            Assert.That(statelessUnitOfWork.IsStarted(), Is.Not.True);
            statelessUnitOfWork.Start();
            statelessUnitOfWork.Commit();
        }

        [Test]
        public void TestWhenNotCommiting()
        {
            Guid guid = Guid.NewGuid();

            var man = new Man(null, null);
            man.Name = "Marcel";
            man.Telephone = guid.ToString();
            man.DateBorn = DateTime.Today.AddYears(-20);

            //Saving new man to DB using a stateless session
            statelessUnitOfWork.Start();
            manRepository.BulkSave(new List<Man>{man});

            statelessUnitOfWork.Commit();

            //Get the man from the DB using his fake (guid) phone number
            unitOfWork.Start();
            Man man2 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.AreEqual("Marcel", man2.Name);

            //Changing the man's name, but this time we "forget" to call t.Commit();
            man2.Name = "Not Marcel";


            //t.Commit(); NOT CALLED
            unitOfWork.Rollback();

            //Get the man from the DB using his fake (guid) phone number
            unitOfWork.Start();
            man2 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.AreEqual("Marcel", man2.Name);

            unitOfWork.Commit();

            //We get rid of the created man
            unitOfWork.Start();
            Man man3 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            manRepository.Delete(man3);

            unitOfWork.Commit();

            //We check it doesn't exist anymore
            unitOfWork.Start();
            Man man4 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.IsNull(man4, "The man is not null");

            unitOfWork.Commit();
        }

        [Test]
        public void TestingConcurrency()
        {
            var threads = new List<Thread>();
            var exceptionCounter = new Counter();
            for (int i = 0; i < 3; i++)
            {
                threads.Add(new Thread(TestingConcurrencyThread));
            }

            foreach (var thread in threads)
            {
                thread.Start(exceptionCounter);
            }

            foreach (var thread in threads)
            {
                thread.Join();
            }

            Assert.That(exceptionCounter.Count, Is.EqualTo(0));
        }

        private void TestingConcurrencyThread(object exceptionCounter)
        {
            try
            {
                unitOfWork.Start();

                Console.Out.WriteLine(humanRepository.GetCount());

                unitOfWork.Commit();
            }
            catch (Exception)
            {
                ((Counter) exceptionCounter).Add();
                throw;
            }
        }

        private class Counter
        {
            private int counter;

            public int Count
            {
                get { return counter; }
            }

            public void Add()
            {
                counter++;
            }
        }
    }
}
