﻿#region Copyright
//       Copyright (c) 2009 - Antoine Jaussoin - http://www.jaussoin.com/blog/
// 
//        This program is free software: you can redistribute it and/or modify
//        it under the terms of the GNU General Public License as published by
//        the Free Software Foundation, either version 3 of the License, or
//        (at your option) any later version.
// 
//        This program is distributed in the hope that it will be useful,
//        but WITHOUT ANY WARRANTY; without even the implied warranty of
//        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//        GNU General Public License for more details.
// 
//        You should have received a copy of the GNU General Public License
//        along with this program.  If not, see <http://www.gnu.org/licenses/>.
#endregion

using System;
using System.Collections.Generic;
using System.Threading;
using FT.Architecture.Example.Core.Entities.Humans;
using FT.Architecture.Example.Core.Entities.Pets;
using NUnit.Framework;

namespace FT.Architecture.Tests.Architecture
{
    [TestFixture]
    public class UnitOfWorkTest : BaseTest
    {
        [Test]
        public void TestInNormalConditions()
        {
            Guid guid = Guid.NewGuid();

            var man = new Man(null, null);
            man.Name = "Bob";
            man.Telephone = guid.ToString();
            man.DateBorn = DateTime.Today.AddYears(-20);

            //Saving new man to DB
            unitOfWork.Start();
            manRepository.Save(man);

            unitOfWork.Commit();

            //Get the man from the DB using his fake (guid) phone number
            unitOfWork.Start();
            Man man2 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.AreEqual("Bob", man2.Name);

            unitOfWork.Commit();

            //We get rid of the created man
            unitOfWork.Start();
            Man man3 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            manRepository.Delete(man3);

            unitOfWork.Commit();

            //We check it doesn't exist anymore
            unitOfWork.Start();
            Man man4 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.IsNull(man4, "The man is not null");

            unitOfWork.Commit();
        }

        [Test]
        public void TestIsStartedWhenStarted()
        {
            unitOfWork.Start();
            Assert.That(unitOfWork.IsStarted(), Is.True);
            unitOfWork.Commit();
        }

        [Test]
        public void TestIsStartedWhenCommitted()
        {
            unitOfWork.Start();
            unitOfWork.Commit();
            Assert.That(unitOfWork.IsStarted(), Is.Not.True);
        }

        [Test]
        public void TestIsStartedWhenRollback()
        {
            unitOfWork.Start();
            unitOfWork.Rollback();
            Assert.That(unitOfWork.IsStarted(), Is.Not.True);
        }

        [Test]
        public void TestIsStartedWhenNotStarted()
        {
            Assert.That(unitOfWork.IsStarted(), Is.Not.True);
            unitOfWork.Start();
            unitOfWork.Commit();
        }

        [Test]
        public void TestWhenNotCommiting()
        {
            Guid guid = Guid.NewGuid();

            var man = new Man(null, null);
            man.Name = "Marcel";
            man.Telephone = guid.ToString();
            man.DateBorn = DateTime.Today.AddYears(-20);

            //Saving new man to DB
            unitOfWork.Start();
            manRepository.Save(man);

            unitOfWork.Commit();

            //Get the man from the DB using his fake (guid) phone number
            unitOfWork.Start();
            Man man2 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.AreEqual("Marcel", man2.Name);

            //Changing the man's name, but this time we "forget" to call t.Commit();
            man2.Name = "Not Marcel";


            //t.Commit(); NOT CALLED
            unitOfWork.Rollback();

            //Get the man from the DB using his fake (guid) phone number
            unitOfWork.Start();
            man2 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.AreEqual("Marcel", man2.Name);

            unitOfWork.Commit();

            //We get rid of the created man
            unitOfWork.Start();
            Man man3 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            manRepository.Delete(man3);

            unitOfWork.Commit();

            //We check it doesn't exist anymore
            unitOfWork.Start();
            Man man4 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.IsNull(man4, "The man is not null");

            unitOfWork.Commit();
        }

        [Test]
        public void TestNewEntityWhenNotCommiting()
        {
            Guid guid = Guid.NewGuid();
            Guid doggyGuid = Guid.NewGuid();

            var man = new Man(null, null);
            man.Name = "Marcel";
            man.Telephone = guid.ToString();
            man.DateBorn = DateTime.Today.AddYears(-20);

            //Saving new man to DB
            unitOfWork.Start();
            manRepository.Save(man);

            unitOfWork.Commit();

            //Get the man from the DB using his fake (guid) phone number
            unitOfWork.Start();
            Man man2 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.AreEqual("Marcel", man2.Name);

            //We create a new doggy, we save it, but we won't commit: it should not end up on the DB
            IPet pet = man2.BuyAPet(doggyGuid.ToString());
            petRepository.Save(pet);

            Assert.AreEqual(1, man2.Pets.Count);

            unitOfWork.Rollback();

            //Get the man from the DB using his fake (guid) phone number
            unitOfWork.Start();
            man2 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.AreEqual("Marcel", man2.Name);

            //We check the doggy is not there in the Pets collection
            foreach (IPet aPet in man2.Pets)
            {
                if (aPet.Name == doggyGuid.ToString())
                    Assert.Fail("The doggy was in the collection");
            }

            Assert.AreEqual(0, man2.Pets.Count);

            //We check the doggy doesn't exist AT ALL
            IPet p = petRepository.GetByMember(d => d.Name, doggyGuid.ToString());
            Assert.IsNull(p, "The doggy shouldn't exist");

            unitOfWork.Commit();

            //We get rid of the created man
            unitOfWork.Start();
            Man man3 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            manRepository.Delete(man3);

            unitOfWork.Commit();

            //We check it doesn't exist anymore
            unitOfWork.Start();
            Man man4 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.IsNull(man4, "The man is not null");

            unitOfWork.Commit();
        }

        [Test]
        public void TestHandledRollback()
        {
            Guid guid = Guid.NewGuid();
            const string name = "Robert";

            var man = new Man(null, null);
            man.Name = name;
            man.Telephone = guid.ToString();
            man.DateBorn = DateTime.Today.AddYears(-20);

            //Saving new man to DB
            unitOfWork.Start();
            manRepository.Save(man);

            unitOfWork.Commit();

            //Get the man from the DB using his fake (guid) phone number
            unitOfWork.Start();
            Man man2 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.AreEqual(name, man2.Name);

            try
            {
                //Potentially dangerous operation, that could (and will) throw an exception
                man2.Name = "Not robert";
                throw new Exception("Wanted Exception");
            }
            catch (Exception)
            {
                //This should abort and not commit any changes
                unitOfWork.Rollback();
            }

            //Get the man from the DB and check the name was NOT changed
            unitOfWork.Start();
            man2 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.AreEqual(name, man2.Name, "The name of the entity changed: " + man2.Name);

            unitOfWork.Commit();

            //We get rid of the created man
            unitOfWork.Start();
            Man man3 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            manRepository.Delete(man3);

            unitOfWork.Commit();

            //We check it doesn't exist anymore
            unitOfWork.Start();
            Man man4 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.IsNull(man4, "The man is not null");

            unitOfWork.Commit();
        }

        [Test]
        public void TestUnhandledRollback()
        {
            Guid guid = Guid.NewGuid();
            const string name = "Gertrude";

            var man = new Man(null, null);
            man.Name = name;
            man.Telephone = guid.ToString();
            man.DateBorn = DateTime.Today.AddYears(-20);

            //Saving new man to DB
            unitOfWork.Start();
            manRepository.Save(man);

            unitOfWork.Commit();

            Man man2;

            try
            {

                //Get the man from the DB using his fake (guid) phone number
                unitOfWork.Start();
                man2 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
                Assert.AreEqual(name, man2.Name);

                //Potentially dangerous operation, that could (and will) throw an exception
                man2.Name = "Not Gertrude";
                throw new Exception("Wanted Exception");
            }
            catch
            {
                unitOfWork.Rollback();
            }

            //Get the man from the DB and check the name was NOT changed
            unitOfWork.Start();
            man2 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.AreEqual(name, man2.Name, "The name of the entity changed: " + man2.Name);

            unitOfWork.Commit();

            //We get rid of the created man
            unitOfWork.Start();
            Man man3 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            manRepository.Delete(man3);

            unitOfWork.Commit();

            //We check it doesn't exist anymore
            unitOfWork.Start();
            Man man4 = manRepository.GetByMember(m => m.Telephone, guid.ToString());
            Assert.IsNull(man4, "The man is not null");

            unitOfWork.Commit();
        }

        [Test]
        public void TestingTheIsTransientMethodOnBaseObject()
        {
            IHuman human = new Man();
            human.Name = "Jean-Robert";
            human.Telephone = "01 12 12 12 12";
            human.DateBorn = DateTime.Today.AddYears(-20);

            Assert.IsTrue(human.IsTransient(), "The object should be declared as being transient at this point");

            unitOfWork.Start();
            humanRepository.Save(human);

            Assert.IsFalse(human.IsTransient(), "The object should not be transient anymore (i.e. persisted)");

            unitOfWork.Commit();

            unitOfWork.Start();
            IHuman reloadedHuman = humanRepository.GetById(human.Id);

            Assert.IsFalse(reloadedHuman.IsTransient(), "The object should not be transient because loaded straight from the DB (i.e. persisted)");
            unitOfWork.Rollback();

            unitOfWork.Start();
            reloadedHuman = humanRepository.GetById(human.Id);

            Assert.IsFalse(reloadedHuman.IsTransient(), "The object should not be transient because loaded straight from the DB (i.e. persisted)");
            humanRepository.Delete(reloadedHuman);
            Assert.IsTrue(reloadedHuman.IsTransient(), "The object should be back to transient: not anymore on the DB, but still there in memory");
            unitOfWork.Rollback();
        }

        [Test]
        public void TestingConcurrency()
        {
            var threads = new List<Thread>();
            var exceptionCounter = new Counter();
            for (int i = 0; i < 3; i++)
            {
                threads.Add(new Thread(TestingConcurrencyThread));
            }

            foreach (var thread in threads)
            {
                thread.Start(exceptionCounter);
            }

            foreach (var thread in threads)
            {
                thread.Join();
            }

            Assert.That(exceptionCounter.Count, Is.EqualTo(0));
        }

        private void TestingConcurrencyThread(object exceptionCounter)
        {
            try
            {
                unitOfWork.Start();

                Console.Out.WriteLine(humanRepository.GetCount());

                unitOfWork.Commit();
            }
            catch (Exception)
            {
                ((Counter) exceptionCounter).Add();
                throw;
            }
        }

        private class Counter
        {
            private int counter;

            public int Count
            {
                get { return counter; }
            }

            public void Add()
            {
                counter++;
            }
        }
    }
}
