﻿using System.Collections.Generic;
using FT.Architecture.Utils.ExtensionMethods;
using NUnit.Framework;

namespace FT.Architecture.Tests.Utils.Extensions
{
    public class EnumerableExtensionsTests
    {
        protected IList<Person> list;

        [SetUp]
        public void SetUp()
        {
            list = new List<Person>();
            list.Add(new Person("Antoine"));
            list.Add(new Person("Anne-Claire"));
            list.Add(new Person("Charlotte"));
        }

        [TestFixture]
        public class When_Joining_An_Enumerable_As_String : EnumerableExtensionsTests
        {
            [Test]
            public void The_Chosen_Member_Should_Join_To_Every_Other_Item()
            {
                // Arrange

                // Act
                string joinedString = EnumerableExtensions.JoinAsString(list, ", ", "Name");

                // Assert
                Assert.That(joinedString, Is.EqualTo("Antoine, Anne-Claire, Charlotte"));
            }

            [Test]
            public void And_The_Member_Doesnt_Exist_It_Should_Throw_An_Exception()
            {
                // Arrange

                // Act & Assert
                Assert.That(() => EnumerableExtensions.JoinAsString(list, ", ", "NonExistantMember"), Throws.Exception);
            }

            [Test]
            public void An_Empty_List_Should_Return_An_Empty_String()
            {
                // Arrange
                list.Clear();

                // Act
                string joinedString = EnumerableExtensions.JoinAsString(list, ", ", "Name");

                // Assert
                Assert.That(joinedString, Is.EqualTo(string.Empty));
            }
        }

        [TestFixture]
        public class When_Taking_A_Subset_Of_A_Collection : EnumerableExtensionsTests
        {
            [Test]
            public void The_Correct_Items_Should_Be_Returned()
            {
                // Arrange

                // Act
                var subset = EnumerableExtensions.Subset(list, 1, 1);

                // Assert
                Assert.That(subset, Has.Count.EqualTo(1));
                Assert.That(subset[0].Name, Is.EqualTo("Anne-Claire"));
            }
        }

        [TestFixture]
        public class When_Building_A_Natural_Sentence_String_From_A_Collection : EnumerableExtensionsTests
        {
            [Test]
            public void The_Correct_Sentence_Should_Be_Returned()
            {
                // Arrange

                // Act
                var sentence = EnumerableExtensions.NaturalSentenceString(list);

                // Assert
                Assert.That(sentence, Is.EqualTo("Antoine, Anne-Claire and Charlotte"));
            }

            [Test]
            public void An_Empty_Collection_Should_Give_An_Empty_String()
            {
                // Arrange
                list.Clear();

                // Act
                var sentence = EnumerableExtensions.NaturalSentenceString(list);

                // Assert
                Assert.That(sentence, Is.EqualTo(string.Empty));
            }

            [Test]
            public void An_Single_Item_Should_Just_Give_That_Single_Item_Back()
            {
                // Arrange
                list.Clear();
                list.Add(new Person("Single Item"));

                // Act
                var sentence = EnumerableExtensions.NaturalSentenceString(list);

                // Assert
                Assert.That(sentence, Is.EqualTo("Single Item"));
            }
        }

        [TestFixture]
        public class When_Extracting_From_A_List : EnumerableExtensionsTests
        {
            [Test]
            public void The_Extracted_Member_On_Each_Item_Should_Be_Returned_As_A_List()
            {
                // Arrange

                // Act
                var extracted = new List<string>(EnumerableExtensions.Extract<string, Person>(list, "Name"));

                // Assert
                Assert.That(extracted, Has.Count.EqualTo(3));
                Assert.That(extracted[0], Is.TypeOf<string>());
                Assert.That(extracted[0], Is.EqualTo("Antoine"));
                Assert.That(extracted[1], Is.EqualTo("Anne-Claire"));
            }

            [Test]
            public void An_Empty_Collection_Should_Give_An_Empty_Collection()
            {
                // Arrange
                list.Clear();

                // Act
                var extracted = new List<string>(EnumerableExtensions.Extract<string, Person>(list, "Name"));

                // Assert
                Assert.That(extracted, Has.Count.EqualTo(0));
            }

            [Test]
            public void A_Non_Existant_Member_Should_Throw_An_Exception()
            {
                // Arrange

                // Act & Assert
                Assert.That(() => EnumerableExtensions.Extract<string, Person>(list, "NonExistantMemeber"), Throws.Exception);
            }
        }
    }

    public class Person
    {
        public Person(string name)
        {
            Name = name;
        }

        public string Name { get; set; }

        public override string ToString()
        {
            return Name;
        }
    }
}
