﻿using FT.Architecture.Utils.ExtensionMethods;
using NUnit.Framework;

namespace FT.Architecture.Tests.Utils
{
    public class StringExtensionsTests
    {
        [TestFixture]
        public class When_Abbreviating_A_String : StringExtensionsTests
        {
            [Test]
            public void A_Too_Long_String_Should_End_By_The_Prefix_And_Be_Of_The_Specified_Size()
            {
                // Arrange
                const string sentence = "Hello World";
                const string suffix = "...";

                // Act
                string result = StringExtensions.Abbreviate(sentence, 5, suffix);

                // Assert
                Assert.That(result, Is.EqualTo("He..."));
            }

            [Test]
            public void A_Short_String_Should_Be_Returned_As_Is()
            {
                // Arrange
                const string sentence = "Hello World";
                const string suffix = "...";

                // Act
                string result = StringExtensions.Abbreviate(sentence, 100, suffix);

                // Assert
                Assert.That(result, Is.EqualTo("Hello World"));
            }

            [Test]
            public void A_String_Of_Exactly_The_Maximum_Size_Should_Be_Returned_As_Is()
            {
                // Arrange
                const string sentence = "Hello World";
                const string suffix = "...";

                // Act
                string result = StringExtensions.Abbreviate(sentence, sentence.Length, suffix);

                // Assert
                Assert.That(result, Is.EqualTo("Hello World"));
            }

            [Test]
            public void A_String_Of_Exactly_The_Maximum_Size_Minus_One_Should_Be_Abbreviated()
            {
                // Arrange
                const string sentence = "Hello World";
                const string suffix = "...";

                // Act
                string result = StringExtensions.Abbreviate(sentence, sentence.Length - 1, suffix);

                // Assert
                Assert.That(result, Is.EqualTo("Hello W..."));
            }

            [Test]
            public void The_Maximum_Size_Can_Never_Be_Greater_Than_The_Suffix_Length()
            {
                // Arrange
                const string sentence = "Hello World";
                const string suffix = "...";

                // Act & Assert
                Assert.That(() => StringExtensions.Abbreviate(sentence, 2, suffix), Throws.Exception);
            }

            [Test]
            public void A_Null_String_Returns_Null()
            {
                // Arrange
                const string suffix = "...";

                // Act
                string result = StringExtensions.Abbreviate(null, 5, suffix);

                // Assert
                Assert.That(result, Is.Null);
            }
        }

        [TestFixture]
        public class When_Trimming_A_String : StringExtensionsTests
        {
            [Test]
            public void Full_Trim_Should_Remove_Any_Non_Alphanumeric_Characters_From_The_Beginning_And_The_End()
            {
                // Arrange
                const string sentence = " §² Hello World! @+ ";

                // Act
                string trimmed = StringExtensions.FullTrim(sentence);

                // Assert
                Assert.That(trimmed, Is.EqualTo("Hello World"));
            }

            [Test]
            public void Start_Trim_Should_Remove_Any_Non_Alphanumeric_Characters_From_The_Beginning()
            {
                // Arrange
                const string sentence = " §² Hello World @+ ";

                // Act
                string trimmed = StringExtensions.FullTrimStart(sentence);

                // Assert
                Assert.That(trimmed, Is.EqualTo("Hello World @+ "));
            }

            [Test]
            public void End_Trim_Should_Remove_Any_Non_Alphanumeric_Characters_From_The_End()
            {
                // Arrange
                const string sentence = " §² Hello World @+ ";

                // Act
                string trimmed = StringExtensions.FullTrimEnd(sentence);

                // Assert
                Assert.That(trimmed, Is.EqualTo(" §² Hello World"));
            }

            [Test]
            public void Full_Trim_Of_Empty_String_Should_Return_Empty_String()
            {
                // Arrange
                string sentence = string.Empty;

                // Act
                string trimmed = StringExtensions.FullTrim(sentence);

                // Assert
                Assert.That(trimmed, Is.EqualTo(string.Empty));
            }

            [Test]
            public void Full_Trim_Of_Spaces_Should_Return_Empty_String()
            {
                // Arrange
                const string sentence = "                 ";

                // Act
                string trimmed = StringExtensions.FullTrim(sentence);

                // Assert
                Assert.That(trimmed, Is.EqualTo(string.Empty));
            }

            [Test]
            public void End_Trim_Of_Spaces_Should_Return_Empty_String()
            {
                // Arrange
                const string sentence = "                 ";

                // Act
                string trimmed = StringExtensions.FullTrimEnd(sentence);

                // Assert
                Assert.That(trimmed, Is.EqualTo(string.Empty));
            }
        }

        [TestFixture]
        public class When_Getting_Friendly_Display : StringExtensionsTests
        {
            [Test]
            public void A_String_With_No_Space_Should_Add_The_Spaces()
            {
                // Arrange
                const string sentence = "ThisIsASentence";

                // Act
                var friendly = StringExtensions.GetFriendlyDisplay(sentence);

                // Assert
                Assert.That(friendly, Is.EqualTo("This Is A Sentence"));
            }

            [Test]
            public void A_String_With_No_Space_And_Numbers_Should_Add_The_Spaces()
            {
                // Arrange
                const string sentence = "ThisIsA123Sentence";

                // Act
                var friendly = StringExtensions.GetFriendlyDisplay(sentence);

                // Assert
                Assert.That(friendly, Is.EqualTo("This Is A 123 Sentence"));
            }

            [Test]
            public void A_String_With_Capitalised_Word_Should_Not_Separate_That_Word()
            {
                // Arrange
                const string sentence = "HeIsFOOBAR";

                // Act
                var friendly = StringExtensions.GetFriendlyDisplay(sentence, EnumerationStyle.UpperAndLowerCaseNoUnderscore);

                // Assert
                Assert.That(friendly, Is.EqualTo("He Is FOOBAR"));
            }

            [Test]
            public void A_String_With_Only_Capitalised_Letter_And_Underscores_Should_Separate_The_Words()
            {
                // Arrange
                const string sentence = "THIS_IS_A_SENTENCE";

                // Act
                var friendly = StringExtensions.GetFriendlyDisplay(sentence, EnumerationStyle.UpperCaseWithUnderscore);

                // Assert
                Assert.That(friendly, Is.EqualTo("This Is A Sentence"));
            }

            [Test]
            public void A_String_With_Only_Capitalised_Letter_And_Numbers_And_Underscores_Should_Separate_The_Words()
            {
                // Arrange
                const string sentence = "THIS_IS_A_123_SENTENCE456";

                // Act
                var friendly = StringExtensions.GetFriendlyDisplay(sentence, EnumerationStyle.UpperCaseWithUnderscore);

                // Assert
                Assert.That(friendly, Is.EqualTo("This Is A 123 Sentence456"));
            }

            [Test]
            public void Getting_Friendly_Display_With_A_Style_That_Doesnt_Exist_Should_Throw_An_Exception()
            {
                // Arrange
                const string sentence = "THIS_IS_A_123_SENTENCE456";

                // Act
                Assert.That(() => StringExtensions.GetFriendlyDisplay(sentence, (EnumerationStyle) 100), Throws.Exception);
            }

            [Test]
            public void An_End_Of_File_Char_Count_As_A_Word_Separation()
            {
                // Arrange
                string sentence = string.Format("ThisIsASent{0}ence", '\0');

                // Act
                var friendly = StringExtensions.GetFriendlyDisplay(sentence);

                // Assert
                Assert.That(friendly, Is.EqualTo("This Is A Sent\0 ence"));
            }

            [Test]
            public void An_Empty_String_Should_Return_Empty_String()
            {
                // Arrange
                string sentence = string.Empty;

                // Act
                string friendlyDisplay = StringExtensions.GetFriendlyDisplay(sentence);

                // Assert
                Assert.That(friendlyDisplay, Is.EqualTo(string.Empty));
            }

            [Test]
            public void An_Empty_String_Should_Return_Empty_String_When_Using_Underscore_Format()
            {
                // Arrange
                string sentence = string.Empty;

                // Act
                string friendlyDisplay = StringExtensions.GetFriendlyDisplay(sentence, EnumerationStyle.UpperCaseWithUnderscore);

                // Assert
                Assert.That(friendlyDisplay, Is.EqualTo(string.Empty));
            }
        }

        [TestFixture]
        public class When_Title_Casing_Each_Word : StringExtensionsTests
        {
            [Test]
            public void A_Lower_Case_Sentence_Should_Capitalise_The_First_Letter_Of_Each_Word()
            {
                // Arrange
                const string sentence = "This is a sentence";

                // Act
                var result = StringExtensions.ToTitleCaseEachWord(sentence);

                // Assert
                Assert.That(result, Is.EqualTo("This Is A Sentence"));
            }

            [Test]
            public void A_Upper_Case_Sentence_Should_Capitalise_The_First_Letter_Of_Each_Word()
            {
                // Arrange
                const string sentence = "THIS IS A SENTENCE";

                // Act
                var result = StringExtensions.ToTitleCaseEachWord(sentence);

                // Assert
                Assert.That(result, Is.EqualTo("This Is A Sentence"));
            }

            [Test]
            public void A_Non_Alpha_Sentence_Should_Return_The_Same()
            {
                // Arrange
                const string sentence = "123;";

                // Act
                var result = StringExtensions.ToTitleCaseEachWord(sentence);

                // Assert
                Assert.That(result, Is.EqualTo("123;"));
            }
        }
    }
}
