﻿#region Copyright
//       Copyright (c) 2009 - Antoine Jaussoin - http://www.jaussoin.com/blog/
// 
//        This program is free software: you can redistribute it and/or modify
//        it under the terms of the GNU General Public License as published by
//        the Free Software Foundation, either version 3 of the License, or
//        (at your option) any later version.
// 
//        This program is distributed in the hope that it will be useful,
//        but WITHOUT ANY WARRANTY; without even the implied warranty of
//        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//        GNU General Public License for more details.
// 
//        You should have received a copy of the GNU General Public License
//        along with this program.  If not, see <http://www.gnu.org/licenses/>.
#endregion
using System;
using System.Collections.Generic;
using System.Diagnostics;
using FT.Architecture.Utils.Collections;
using NUnit.Framework;

namespace FT.Architecture.Tests.Utils
{
    /// <summary>
    /// Tests for the Hashed List
    /// </summary>
    [TestFixture]
    public class HashedListTest
    {
        private const int TEST_LIST_SIZE = 100000;
        private static readonly Random _random = new Random((int)DateTime.Now.Ticks);

        [Test]
        public void TestAddRemove()
        {
            HashedList<DateTime> dates = new HashedList<DateTime>();

            DateTime date1 = new DateTime(2010,1,1);
            DateTime date2 = new DateTime(2009,1,1);
            DateTime date3 = new DateTime(2008,1,1);

            dates.Add(date1);
            Assert.AreEqual(1, dates.Count, "Count should be one");

            DateTime d = dates[0];
            Assert.AreEqual(date1, d, "Dates should be the same");

            dates.RemoveAt(0);
            Assert.AreEqual(0, dates.Count);

            bool succeeded = dates.Remove(date1);
            Assert.IsFalse(succeeded);

            dates.Add(date1);
            dates.Add(date3);
            dates.Insert(1, date2);

            Assert.AreEqual(dates[0], date1);
            Assert.AreEqual(dates[1], date2);
            Assert.AreEqual(dates[2], date3);
            Assert.AreEqual(3, dates.Count);

            dates.Remove(date2);
            Assert.AreEqual(dates[0], date1);
            Assert.AreEqual(dates[1], date3);
            Assert.AreEqual(2, dates.Count);


            Assert.Throws(typeof (ArgumentException), () => dates.Add(date1));

            dates.Clear();
            Assert.AreEqual(0, dates.Count);

            dates.Add(date1);
            dates.Add(date3);
            dates.Insert(1, date2);
            Assert.IsTrue(dates.Contains(date1));
            Assert.IsTrue(dates.Contains(date2));
            Assert.IsTrue(dates.Contains(date3));

            Assert.AreEqual(0, dates.IndexOf(date1));
            Assert.AreEqual(1, dates.IndexOf(date2));
            Assert.AreEqual(2, dates.IndexOf(date3));
        }

        [Test]
        public void TestPerformanceContains()
        {
            IList<DateTime> dates = GetFullList(TEST_LIST_SIZE);
            
            DateTime fakeDate = new DateTime(1900,1,1);

            Stopwatch sw = new Stopwatch();
            sw.Start();

            for (int i = 0; i < TEST_LIST_SIZE; i++)
            {
                if (dates.Contains(fakeDate))
                    throw new Exception("Shouldn't happen");
            }

            DateTime existingDate = dates[(dates.Count * 2) / 3];

            for (int i = 0; i < TEST_LIST_SIZE; i++)
            {
                if (!dates.Contains(existingDate))
                    throw new Exception("Shouldn't happen");
            }

            sw.Stop();

            Console.Out.WriteLine("Took "+sw.ElapsedMilliseconds);
            Assert.Less(sw.ElapsedMilliseconds, 500, "This should take less than 500ms, took "+sw.ElapsedMilliseconds);

        }

        [Test]
        public void TestPerformanceIndexer()
        {
            IList<DateTime> dates = GetFullList(TEST_LIST_SIZE);

            int index = (dates.Count*2)/3;

            Stopwatch sw = new Stopwatch();
            sw.Start();

            for (int i = 0; i < TEST_LIST_SIZE; i++)
            {
                DateTime date = dates[index];
            }

            sw.Stop();

            Console.Out.WriteLine("Took " + sw.ElapsedMilliseconds);
            Assert.Less(sw.ElapsedMilliseconds, 500, "This should take less than 500ms, took " + sw.ElapsedMilliseconds);

        }

        private IList<DateTime> GetFullList(int size)
        {
            
            IList<DateTime> list = new HashedList<DateTime>();

            //Generates fake data
            for (int i = 0; i < size; i++)
            {
                DateTime date = new DateTime(2010, 1, 1).AddMilliseconds(_random.Next(0, int.MaxValue));
                if (!list.Contains(date))
                    list.Add(date);
            }

            return list;
        }
    }
}
